/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/*This is the AEP test application. It is used to verify that the product has been 
correctly installed and is functioning properly. It has two modes: standard (with no command line
arguments) and verbose mode (with the -v command line argument). 
The application 
1) Generates a number of 2500 byte hardware random numbers and carries out a FIPS check on their randomness
2) Generates  a number of 2500 byte pseudo-random numbers and carries out a FIPS check on their randomness
3) Checks the results of an AEP Modular Exponentiation without CRT and an AEP Modular Exponentiation with CRT
   against the known answer and the result of a software Mod Exp (Code from Open SSL).
4) Calculates the rate of Mod Exp calculations.

Standard mode prints short messages to the user. Verbose mode is used to give the user more information 
about what's going on, especially in the case of an error. In verbose mode the program's version 
number is also printed*/


#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#include "aep_defs.h"
#include "aep_api.h"
#include "aeptest.h"
#include "montgomery.h"
#include "randtest.h"
#include "aeptest_version.h"
#include "aepmalloc.h"


int
main(int argc, char *argv[])
{
	AEP_RV rv;
	struct rlimit rlim;
	AEP_CONNECTION_HNDL firstConnection;
	
	/*Inputs for the Mod Exp tests*/
	BIGNUM 	*base, *power, *mod, *p, *q, *dmpi, *dmqi, *iqmp;
	
	/*Answers from the ModExp correctness tests*/
	BIGNUM *softwareResult, *hardwareNoCRT, *hardwareWithCRT, *knownResult;
	
	/*Pointer to the random number that we will generate*/
	void *randNum;
	
	/*Timing variables*/
	struct timeval startTime, stopTime, diffTime;
	double  doubleTime,
		modExpRate;
	
	/*Mod Exp threads*/
	pthread_t pthrThread;
	
	/*String to hold version number*/
	AEP_CHAR versionString[100];
	
	/*Variables to keep track of errors*/
	int randErrorCount= 0;
	int thisRandErrorCount= 0;
	int modExpCorrectnessErrors;
	int err;
	int returnVal;
	/*This variable will keep track of the rv of any errors that occur*/
	AEP_U32 errCode=0;
	
	/*Counter variable*/
	int counter=0x11223344;
	long i;
	
	
	
	
	/*Check whether the -v option has been given on the command line*/
	for(counter = 1; counter < argc; counter ++)
	{
		if (strcmp (argv[counter], VERBOSE_OPTION)==0)
			verboseMode = 1;
		else
		{
			printf("Illegal option: %s \nCorrect usage: aeptest [-v]\n", argv[counter]);
			return (1);
		}
	}
	
	
	/*If the program is run in verbose mode, the first thing we do is print the version number of this test*/
	if(verboseMode)
	{
		strcpy(versionString, AEPTEST_VERSION);
		printf("AEP Test (version %s %s)\n", versionString, COMPILER_VERSION);
	}
	
	getrlimit(RLIMIT_NOFILE,&rlim);
	rlim.rlim_cur=rlim.rlim_max;
	setrlimit(RLIMIT_NOFILE,&rlim);
	
	
	rv = AEP_Initialize(NULL);
	/*Need to do a lot of error checking here, because if the daemon isn't running 
	or has a problem it will show up here for the first time*/
	/*Check whether this was successful*/
	if(rv != AEP_R_OK)
	{
		printf("Error initializing the API: ");
		printErrorMessage(rv); 	
		/*In this case a GENERAL ERROR will be returned if you cannot connect to the daemon*/ 		 		
		if(rv == AEP_R_GENERAL_ERROR)  		
			printf("\n(Is the daemon running?)\n"); 				
		return(rv);
	}
	
	/*Set up the attributes for the threads*/
	pthread_attr_init(&attr1);
	pthread_attr_setdetachstate(&attr1, PTHREAD_CREATE_DETACHED);
	pthread_attr_setscope(&attr1, PTHREAD_SCOPE_SYSTEM);
	
	AEP_SetBNCallBacks(&GetBigNumSize, &MakeAEPBigNum, &ConvertAEPBigNum);
	
	
	rv = AEP_OpenConnection(&firstConnection);
	if (rv != AEP_R_OK)
	{
		printf("Error! Could not open a connection to the daemon.\n");
		if(verboseMode == 1)
		{                       
			/*Check all the different return values to see what error occurred*/
			printErrorMessage(rv);
			printf("\n");
		}
		return(rv);
	}  
	
	
	if(NUM_HARDWARE_RAND_TESTS > 0)
	{
		/*Check the GenRandom function*/
		/*Note: Generate random works in two modes. The first generates pure hardware random numbers and 
		you specify this by setting the type argument to 1. The second generates pseudo-random numbers
		(seeded by hardware) and you specify it by setting type to 2. We need to test both versions*/
		printf("RNG test (hardware)...");
		/*Flush the standard output buffer, otherwise this string will not be 
	  	displayed straight away*/
		fflush( stdout );
		
		/*Allocate space for the new random number*/
		randNum = (void *)AEP_malloc(2500);
		if(randNum==NULL)
		{
			printf("\nError allocating host memory...exiting\n");
			return(1);
		}
		
		/*Note: we need to generate 2500 bytes of random data for the FIPS randomness 
		 test. When testing the hardware random numbers, we generate this in chunks, 
		 so that we don't put too much pressure on the card*/
		for(counter = 0; counter < NUM_HARDWARE_RAND_TESTS; counter ++)
		{
			thisRandErrorCount = 0;
			for (i=0; (i<(2500 / HARDWARE_RANDOM_CHUNKS)); i++)
			{
				rv = AEP_GenRandom(firstConnection, HARDWARE_RANDOM_CHUNKS, HARDWARE_RANDOM, (void *)((long)randNum + i*HARDWARE_RANDOM_CHUNKS), NULL);
				if(rv != AEP_R_OK)
				{
					randErrorCount ++;
					thisRandErrorCount++;
					errCode = rv;
					
					if(verboseMode == 1)
					{
						printf("\nError generating random number: ");
						printErrorMessage(rv);
					}/*End if (verboseMode ==1)*/
					
				}/*End if(rv!=AEP_R_OK)*/
				
				
			}/*End For*/
			if(thisRandErrorCount ==0)
			{
				/*Put the data through a FIPS randomness test to show that
				  it really is random*/
				err = randTest((unsigned char *)randNum);
				if(err !=0)
				{
					if(verboseMode == 1)
						printf("\nError! Failed randomness test");
					errCode = HARDWARE_RAND_INCORRECT;
					randErrorCount++;
				}
			}
		}/*End for*/ 
		
		/*Print a message to let the user know we completed successfully*/ 
		if(randErrorCount == 0)
			printf("ok");
		else if(verboseMode == 0)
			printf("error!");
		
		if(verboseMode ==1)
		{	
			printf("\n(%d %d-byte random number", NUM_HARDWARE_RAND_TESTS, 2500);
			if(NUM_HARDWARE_RAND_TESTS >1)
				printf("s");
			printf(" generated. %d error",  randErrorCount);	
			if(randErrorCount != 1)
				printf("s");
			printf(" occurred)");
		}
		
		/*Free up memory*/
		AEP_free(randNum);

	}/*End if (NUM_HARDWARE_RANDS > 0)*/
	
	
	/*Test the pseudo random numbers*/
	if(NUM_SOFTWARE_RAND_TESTS > 0)
	{
		
		randErrorCount = 0;
		
		
		/*Allocate space for the new random number*/
		randNum = (void *)AEP_malloc(2500);
		if(randNum==NULL)
		{
			printf("\nError allocating host memory...exiting\n");
			return(1);
		}

		printf("\nRNG test (pseudo-random)...");
		/*Flush the standard output buffer, otherwise this string will not be 
	  	displayed straight away*/
		fflush( stdout );
		for(counter = 0; counter < NUM_SOFTWARE_RAND_TESTS; counter ++)
		{
			thisRandErrorCount = 0;
			/*Note: these numbers should be 2500 bytes long, since that's what the randomness test takes*/
			rv = AEP_GenRandom(firstConnection, 2500, PSEUDO_RANDOM, randNum, NULL);
				
			if(rv != AEP_R_OK)
			{
				randErrorCount++;
				thisRandErrorCount++;
				errCode = rv;
				
				if(verboseMode == 1)
				{
					printf("\nError generating pseudo random number: ");
					printErrorMessage(rv);
				}/*End if (verboseMode == 1)*/
			}/*End if(rv!=AEP_R_OK)*/
			
			if(thisRandErrorCount ==0)
			{
				err = randTest((unsigned char *)randNum);
				if(err !=0)
				{
					errCode = PSEUDO_RAND_INCORRECT;
				
					if(verboseMode == 1)
						printf("\nError! Failed randomness test");
					randErrorCount++;
				}
			}
			
		}/*End for*/
		
		if(randErrorCount == 0)
			printf("ok");
		else if(verboseMode == 0)
			printf("error!");
		
		if(verboseMode == 1)
		{	
			printf("\n(%d %d-byte pseudo-random number", NUM_SOFTWARE_RAND_TESTS, 2500);
			if(NUM_SOFTWARE_RAND_TESTS >1)
				printf("s");
			printf(" generated. %d error",  randErrorCount);
			if(randErrorCount != 1)
				printf("s");
			printf(" occurred) \n");
		}
		else
			printf("\n");

		/*Free up memory*/
		AEP_free(randNum);
	}/* end if(NUM_SOFTWARE_RAND_TESTS > 0)*/
	
	
	/*Do functional test on modExps*/
	/*To check the correctness of the functions we use data that we know the answer to.
	Then we do an AEP mod exp (without CRT), an AEP mod exp with CRT, a software mod exp 
	(without CRT, code from OpenSSL), and compare the result from all of these to the 
	known answer*/
	
	printf("Modular exponentiation correctness test...");
	/*Flush the standard output buffer, otherwise this string will not be 
	  displayed straight away*/
	fflush( stdout );
	
	/*Set up the data for the modExps (This is all hardcoded in Montgomery.c)*/
	base = BN_SetCorrectnessBase();
	power = BN_SetCorrectnessPower();
	mod = BN_SetCorrectnessMod();
	p = BN_SetCorrectnessP();
	q = BN_SetCorrectnessQ();
	dmpi = BN_SetCorrectnessDmpi();
	dmqi = BN_SetCorrectnessDmqi();
	iqmp = BN_SetCorrectnessIqmp();
	knownResult = BN_SetCorrectnessResult();
	
	
	softwareResult=BN_new();
	BN_init(softwareResult);
	hardwareNoCRT=BN_new();
	BN_init(hardwareNoCRT);
	hardwareWithCRT=BN_new();
	BN_init(hardwareWithCRT);
	
	modExpCorrectnessErrors=0;
	
	
	/*Calculate the AEP result (no CRT)*/
	rv = AEP_ModExp(firstConnection, base, power, mod, hardwareNoCRT, NULL);		
	if(rv != AEP_R_OK)
	{
		errCode = rv;
		modExpCorrectnessErrors ++;
		if(verboseMode ==1)
		{
			printf("\nError calculating AEP Mod Exp: ");
			printErrorMessage(rv);
		}
	}
	else
	{
		rv = BN_ucmp(hardwareNoCRT, knownResult);
		if(rv!=0)
		{
			errCode = AEP_MOD_EXP_INCORRECT;
			modExpCorrectnessErrors++;
			if(verboseMode ==1)
				printf("\nError! The result of the AEP modular exponentiation (without CRT) was not correct.");
			
		}
	}
	
	/*Calculate the AEP result with CRT*/
	rv = AEP_ModExpCrt(firstConnection, base, p, q, dmpi, dmqi, iqmp, hardwareWithCRT, NULL);
	if(rv != AEP_R_OK)
	{
		errCode = rv;
		modExpCorrectnessErrors ++;
		if(verboseMode ==1)
		{
			printf("\nError calculating AEP Mod Exp CRT: ");
			printErrorMessage(rv);
		}
	}
	else
	{
		rv = BN_ucmp(hardwareWithCRT, knownResult);
		if(rv!=0)
		{
			errCode = AEP_MOD_EXP_CRT_INCORRECT;
			modExpCorrectnessErrors++;
			if(verboseMode ==1)
				printf("\nError! The result of the AEP modular exponentiation (with CRT) was not correct.");
			
		}
	}
	
	
	/*Calculate the software result. Note: this function returns 1 on completion, or 0 on error*/
	/*Only do this check if the results above were correct*/
	if(modExpCorrectnessErrors ==0)
	{
		rv = modExp(softwareResult, base, power, mod);
		if(rv!=1)
		{
		
			errCode = 1;
			modExpCorrectnessErrors++;
			if(verboseMode ==1)
				printf("\nError calculating Mod Exp in software.");
			
		}
		else{
			/*Check the software result against our harcoded result to verify correctness*/
			rv = BN_ucmp(softwareResult, knownResult);
			if(rv!=0)
			{
				errCode = SOFTWARE_MOD_EXP_INCORRECT;
				modExpCorrectnessErrors++;
				
				if(verboseMode ==1)
					printf("\nError! The hardware results do not match software calculations.");
			}
		}
	}

	if(modExpCorrectnessErrors==0)
		printf("ok");
	else if(verboseMode == 0)
		printf("error!");
		
	
	
	/*Disconnect firstConnection*/
	rv = AEP_CloseConnection(firstConnection);
	if (rv != AEP_R_OK)
	{
		printf("Error closing connection to the daemon \n");
		printErrorMessage(rv);
		printf("\n");
		return(rv);
	}  
	
	
	/*Free up memory*/
	BN_clear_free(base);
	BN_clear_free(power);
	BN_clear_free(mod);
	BN_clear_free(p);
	BN_clear_free(q);
	BN_clear_free(dmpi);
	BN_clear_free(dmqi);
	BN_clear_free(iqmp);
	BN_clear_free(knownResult);
	BN_clear_free(softwareResult);
	BN_clear_free(hardwareNoCRT);
	BN_clear_free(hardwareWithCRT);
				
	
	/*Calculate the rate of Mod Exps*/
	/*We test this by starting a number of threads, each of which executes a number of modular 
	exponentiations. We check how long it takes the threads to finish their work and calculate
	the rate from this*/ 
	if(NUM_MOD_EXPS > 0)
	{
		printf("\nModular exponentiation rate, please wait...");
		fflush(stdout);
		for(counter=0;counter<NUMTHREADS;counter++)
		{
			rv = AEP_OpenConnection(&hConnection[counter]);
			if (rv != AEP_R_OK)
			{
				printf("Error! Could not open a connection to the daemon. ");
				errCode = rv;
				
				if(verboseMode == 1)
				{
					/*Check all the different return values to see what error occurred*/
					printErrorMessage(rv);
					printf("\n");
				}
				
				/*Close all connections opened up to now*/
				for(i=0; i<counter; i++)
				{
					rv = AEP_CloseConnection(hConnection[counter]);
				}
				return(rv);
			}    
		}
		
		
		/*Take the start time here */
		returnVal = gettimeofday(&startTime, NULL);
		if(returnVal != 0)
		{
			printf("Error reading system time...exiting\n\n");
			for(i=0; i<NUMTHREADS; i++)
			{
				rv = AEP_CloseConnection(hConnection[i]);
			}
			return(1);
		}
		
		for(counter=0;counter<NUMTHREADS;counter++)
		{       
			/*Create the Mod Exp threads*/
			if (pthread_create(&pthrThread, &attr1, ptrModExpThread, (void *) &hConnection[counter]))
			{	
				fprintf(stderr,"Cannot create threads...exiting\n");
				return(1);
			}
		}
		
		pthread_mutex_lock(&count_lock);
		
		while(finished==0)
			pthread_cond_wait(&ThreadsFinished, &count_lock);
		pthread_mutex_unlock(&count_lock);
		
		
		returnVal = gettimeofday(&stopTime, NULL);
		if(returnVal != 0)
		{
			printf("Error reading system time...exiting\n\n");
			for(i=0; i<NUMTHREADS; i++)
			{
				rv = AEP_CloseConnection(hConnection[i]);
			}
			return(1);
		}
		
		
		/*Subtract the startTime from the stopTime*/
		if(stopTime.tv_usec < startTime.tv_usec)
		{
			stopTime.tv_usec += 1000000;
			stopTime.tv_sec -=1;
		}
		diffTime.tv_usec = stopTime.tv_usec - startTime.tv_usec;
		diffTime.tv_sec = stopTime.tv_sec - startTime.tv_sec;
		
		/*Error checks*/
		if ((diffTime.tv_usec <0)||(diffTime.tv_sec < 0))
		{
			printf("Error reading system time...exiting\n\n");
			for(i=0; i<NUMTHREADS; i++)
			{
				rv = AEP_CloseConnection(hConnection[i]);
			}
			return(1);
		}
		
		doubleTime = diffTime.tv_sec + (diffTime.tv_usec / (double)1000000);
		/*Error check*/
		if (doubleTime <=0)
		{
			printf("Error reading system time...exiting\n\n");
			for(i=0; i<NUMTHREADS; i++)
			{
				rv = AEP_CloseConnection(hConnection[i]);
			}
			return(1);
		}
		
		
		/*Check whether you finished the test successfully, or exceeded the maximum no of errors*/
		if(modExpErrorCount< ERROR_LIMIT)
		{
			/*Now calculate the average no. of mod exp's per second*/
			modExpRate = (NUM_MOD_EXPS * NUMTHREADS) / doubleTime;
			printf("%.3f per second (1024-bit)\n", modExpRate);
			if(verboseMode ==1)
			{
				printf("(Executed %d modular exponentiations in %.3f seconds, with %d errors) \n", NUMTHREADS * NUM_MOD_EXPS, doubleTime, modExpErrorCount);
			}
			
		}
		else
		{
			/*Tell the user that you have reached the error limit*/
			printf("\nExceeded the maximum no. of errors...exiting\n\n");
			errCode = EXCEEDED_MOD_EXP_RATE_ERROR_LIMIT;
		}
		
		
		if((modExpErrorCount >0)||(randErrorCount>0)||(modExpCorrectnessErrors>0))
		{
			printf("\n********************* ERRORS HAVE OCCURRED *********************\n");
			printf("For further information on errors please turn on tracing (see documentation) and check the AEP log files\n\n");  	
		}
		 		 	
		for(counter=0;counter<NUMTHREADS;counter++) 		
		{
			rv = AEP_CloseConnection(hConnection[counter]);
			if (rv != AEP_R_OK)
			{
				printf("Could not disconnect for connection %d\n", counter);
				
				if(verboseMode == 1)
				{
					if(rv == AEP_R_AEPAPI_NOT_INITIALIZED)
						printf("Error closing connection to the daemon: API not initialized\n");
				}
				
				return(rv);
			}    
		}
		
		}/*End if (NUM_MOD_EXPS > 0)*/
		
		
		rv = AEP_Finalize();
		if (rv != AEP_R_OK)
		{
			printf("Error finalizing ");
			if(rv == AEP_R_AEPAPI_NOT_INITIALIZED)
				printf(": API not initialized\n");
			return(rv);
		} 
		
		
		return(errCode);
		
		
}

/*This function does the work of calculating the rate of Mod Exps*/
static void *ptrModExpThread(void *arg)
{
	AEP_RV rv;
	AEP_CONNECTION_HNDL hConnection;
	BIGNUM *a, *p, *q, *r, *dmpi, *dmqi, *iqmp;
	int i;
	
	hConnection = *(AEP_CONNECTION_HNDL*)(arg);
	
	a = BN_SetBase();
	p = BN_SetPower();
	q = BN_SetMod();
	
	dmpi = BN_SetMod();
	dmqi = BN_SetPower();
	iqmp = BN_SetBase();
	

	p->top = p->max = 16;
	q->top = q->max = 16;
	dmpi->top = dmpi->max = 16;
	dmqi->top = dmqi->max = 16;
	iqmp->top = iqmp->max = 16;
	
	
	
	for (i=0;i<NUM_MOD_EXPS;i++)
	{
		r=BN_new();
		BN_init(r);
		
		rv = AEP_ModExpCrt(hConnection, a, p, q, dmpi, dmqi, iqmp, r, NULL);
		if (rv==AEP_R_OK)
		{
			BN_clear_free(r);               
		}
		else
		{
			modExpErrorCount++;
			if(verboseMode ==1)
			{
				printf("\nError calculating AEP Mod Exp CRT: ");
				printErrorMessage(rv);
			}
		}
		
		/*Check whether you reached the error limit - if so stop the thread*/
		if(modExpErrorCount >= ERROR_LIMIT)
		{       
			count_exits();
			pthread_exit(NULL);
		}
		
	}
	
	
	BN_clear_free(a);
	BN_clear_free(p);
	BN_clear_free(q);
	BN_clear_free(dmpi);
	BN_clear_free(dmqi);
	BN_clear_free(iqmp);
	
	count_exits();
	
	pthread_exit(NULL);

	return ((void *)NULL);
	
}

/*Check whether all threads have finished*/
static void count_exits()
{
	static int count = 0;
	pthread_mutex_lock(&count_lock);
	count++;
	if(count==NUMTHREADS)
    {
        finished=1;
		pthread_mutex_unlock(&count_lock);
		pthread_cond_signal(&ThreadsFinished);
    }
	else
        pthread_mutex_unlock(&count_lock);
	
}



/*Callback functions*/
static AEP_RV GetBigNumSize(void* ArbBigNum, AEP_U32* BigNumSize)
{
	BIGNUM* bn;
	
	/*Cast the ArbBigNum pointer to our BIGNUM struct*/
	bn = (BIGNUM*) ArbBigNum;
	
	/*Size of the bignum in bytes is equal to the bn->top (no of 32 bit words) multiplies by 4*/
	*BigNumSize = bn->top << 2;
	
	return AEP_R_OK;
}

static AEP_RV MakeAEPBigNum(void* ArbBigNum, AEP_U32 BigNumSize, unsigned char* AEP_BigNum)
{
	BIGNUM* bn;
	unsigned char* buf;
	int i;
	
	/*Cast the ArbBigNum pointer to our BIGNUM struct*/
	bn = (BIGNUM*) ArbBigNum;
	
	/*Must copy data into a (monotone) least significant byte first format
	Note this is a Solaris box so data is stored in Big Endian format, but our big num format
	is least significant 32 bit block first.  So we need to swap the bytes in the words*/
	for(i=0;i<bn->top;i++)
	{
		buf = (unsigned char*)&bn->d[i];
		
		*((AEP_U32*)AEP_BigNum) = (AEP_U32) ((unsigned) buf[1] << 8 | buf[0]) |
			((unsigned) buf[3] << 8 | buf[2])  << 16;
		
		AEP_BigNum += 4;
	}
	return AEP_R_OK;
}

/*Turn an AEP Big Num back to a user big num*/
static AEP_RV ConvertAEPBigNum(void* ArbBigNum, AEP_U32 BigNumSize, unsigned char* AEP_BigNum)
{
	BIGNUM* bn;
	int i;
	
	bn = (BIGNUM*)ArbBigNum;
	
	if (BigNumSize % 4 != 0)
	{
		printf("Must be multiple of 4 bytes");
		exit(1);
	}
	bn->top = bn->max = BigNumSize >> 2;
	
	bn->d = (unsigned long*)AEP_malloc(BigNumSize);
	
	for(i=0;i<bn->top;i++)
	{
		bn->d[i] = (AEP_U32) ((unsigned) AEP_BigNum[3] << 8 | AEP_BigNum[2]) << 16 |
			((unsigned) AEP_BigNum[1] << 8 | AEP_BigNum[0]);
		AEP_BigNum += 4;
	}
	
	return AEP_R_OK;
}

static void printErrorMessage(AEP_RV returnVal)
{
	switch(returnVal){
		case AEP_R_GENERAL_ERROR:
			printf("General error");
			break;
		case AEP_R_HOST_MEMORY:
			printf("Host memory error");
			break;
		case AEP_R_FUNCTION_FAILED:
			printf("Function failed");
		case AEP_R_ARGUMENTS_BAD:
			printf("Invalid arguments to message");
			break;
		case AEP_R_NO_TARGET_RESOURCES:
			printf("Insufficient resources on target to complete request");
			break;
		case AEP_R_SOCKERROR:
			printf("Socket error");
			break;
		case AEP_R_SOCKEOF:
			printf("Socket EOF");
			break;
		case AEP_R_CONNECTION_HANDLE_INVALID:
			printf("Invalid connection handle");
			break;
		case AEP_R_TRANSACTION_HANDLE_INVALID:
			printf("Invalid transaction handle");
			break;
		case AEP_R_TIMED_OUT:
			printf("Timed out");
			break;
		case AEP_R_FXN_NOT_IMPLEMENTED:
			printf("Function not implemented");
			break;
		case AEP_R_TARGET_ERROR:
			printf("Target Error");
			break;
		case AEP_R_DAEMON_ERROR:
			printf("Daemon Error");
			break;
		case AEP_R_MUTEX_BAD:
			printf("Error obtaining a mutex");
			break;
		case AEP_R_AEPAPI_NOT_INITIALIZED:
			printf("API not initialized");
			break;
		case AEP_R_NO_MORE_CONNECTION_HNDLS:
			printf("Unable to get connection handle");
			break;
		case AEP_R_AEPAPI_ALREADY_INITIALIZED:
			printf("API already initialized");
			break;
		default:
			printf("Error code: %0x ", returnVal);
			break;
	}
}


