/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

#include "boot.h"

/************************************************/
/* Global Variables								*/
int bVerbose;
int bDoVersion;
int bDoPause;
int bReset;

int dev_fd;					/* AEP crypto accelerator file descriptor */
FILE *img_fd;

char acAEPDeviceName[100];	/* the name of the file associated with dev_fd, eg /dev/paep0 */
char acAEPImageName[100];	/* the name of the file with the target to load */



/*******************************************************************************/
/*	main()

	Downloads an image to the target.  The image is written sequentially to the 
	target starting at the load address.  A CRC is calculated which is then verified
	by the target and if orrect, the entry address given will be used to begin
	execution on the target.

	Input:	
	-device		Specifies the destination device
	-image		Specifies the source binary image file
	-entry		Specifies the address in memory for the target to begin execution
	after download
	-load		Specifies the beginning destination address for the image on the target
	-reset		Send a reset command before starting the image download.
	-verbose	Turns on debug messages
	-pause		Waits for keypress after each write and read

	Output:
	int			- Zero if Successful, negative if an error occurred

*/
int main(int argc, char** argv)
{
	int iProceed=0;			/* 1 if REQUIRED fields device and image are specified, 0 otherwise */
	AEP_U32 iTargetEntryAddress = AEP_BOOT_IMG_ADDRESS;
	AEP_U32 iTargetLoadAddress = AEP_BOOT_IMG_ADDRESS;
	int iRetVal;

	/* Initialize globals */
	bVerbose = 0;
	bDoVersion = 0;
	bDoPause = 0;
	bReset = 0;
	dev_fd = 0;				/* AEP crypto accelerator file descriptor */
	img_fd = (FILE *)NULL;

	/****************** parse args ********************/
	
	argc--;
	argv++;
	iProceed = 0;
	
	while(argc>=1){
		
		if(strcmp(*argv, "-device") == 0){
			if(--argc < 1)
				return usage();
			strcpy(acAEPDeviceName,*(++argv));
			iProceed |= 0x01;
		}
		
		if(strcmp(*argv, "-image") == 0){
			if(--argc < 1)
				return usage();
			strcpy(acAEPImageName,*(++argv));
			iProceed |= 0x02;
		}
		
		if(strcmp(*argv, "-load") == 0){
			if(--argc < 1)
				return usage();
			sscanf(*(++argv), "%X", &iTargetLoadAddress );
		}
		
		if(strcmp(*argv, "-entry") == 0){
			if(--argc < 1)
				return usage();
			sscanf(*(++argv), "%X", &iTargetEntryAddress );
		}
		
		if(strcmp(*argv, "-version") == 0)
			bDoVersion=1;

		if(strcmp(*argv, "-reset") == 0)
			bReset=1;
		
#ifdef THIS_IS_NOT_FOR_RELEASE /* obviously, the customer should not have access to this: */
		
		if(strcmp(*argv, "-verbose") == 0)
			bVerbose=1;
		
		if(strcmp(*argv, "-pause") == 0)
			bDoPause=1;
#endif
		
		argc--;
		argv++;
	}
	
	/*********** Display loader version info ***********/
	if(bDoVersion){
		printf("AEP PCI Loader: Version %s %s \n",LOADER_VERSION, COMPILER_VERSION);
		return RV_VERSION_DISPLAYED;
	}
	
	/* Device filename and image filename MUST be provided */
	if(iProceed != 0x03){
		/*Display required param info */
		return usage();
	}
	
	/**************** open /dev/paep0 ******************/
	dev_fd = OpenAccel( acAEPDeviceName );

	iRetVal = 0;
		
	if (dev_fd < 0) {
		return RV_DEVICE_ERR;
	}
	else if ( bReset ){
		/* Send the reset command to the target in case an application lives there already */
		iRetVal = ResetDev( dev_fd );

		switch ( iRetVal ) {
			case RV_WRITE_ERROR:
				fprintf( stderr, "ERROR: Write Failed.  Reset unsuccessful : %s\n",
						strerror(errno));
				break;

			case RV_READ_ERROR:
				fprintf(stderr, "ERROR: Read failed.  Reset unsuccessful : %s\n",
						strerror(errno));
				break;	
		}
	}

	/**************** open image file ******************/	 
	if ( iRetVal == 0 ) {
		/* if Reset command was successful or did not take place */
		img_fd = OpenImage( acAEPImageName );
	
		if (img_fd != NULL) {
			/* Image file has been found, now download the image to the target and boot it */
			iRetVal = BootTarget( dev_fd, img_fd, iTargetLoadAddress, iTargetEntryAddress );

			switch ( iRetVal ) {
				case RV_APP_ALREADY_LOADED:
					fprintf( stderr, "ERROR: Download Failed.  Application already loaded.\n" );
					break;

				case RV_WRITE_ERROR:
					fprintf( stderr, "ERROR: Write Failed.  Download unsuccessful : %s\n",
							strerror(errno) );
					break;

				case RV_READ_ERROR:
					fprintf(stderr, "ERROR: Read failed.  Target device is not booted : %s\n",
							strerror(errno) );
					break;	
			}
		}
		else {
			/* There was a problem finding or opening the image source file.*/
			iRetVal=RV_IMAGE_ERR;
		}
	}

	/* Close device before exiting */
	if(dev_fd){
		if(bVerbose) printf("Closing %s ...", acAEPDeviceName);
		close(dev_fd);
		if(bVerbose) printf("Done\n");
	}
	
	/* Close image file before exiting */
	if(img_fd){
		if(bVerbose) printf("Closing %s ...", acAEPImageName);
		fclose(img_fd);
		if(bVerbose) printf("Done\n");
	}
	
	return iRetVal;
}



/*******************************************************************************/
/*	ResetDev( ... )

	Sends a reset command to the selected destination device.  After sending the
	write and reading the response, a small delay is used to allow the device to
	reset and re-init.

	Input:	
	dev_fd		- The destination device file descriptor (target)

	Output:
	int			- Zero if Successful, non-zero if an error occurred

*/
int ResetDev( int iDevFD )
{
	AEP_U8 acReply[AEP_BLOCK_SIZE]; /* the reply buffer (when target replies to a command) */
	AEP_U8 acCmd[AEP_BLOCK_SIZE];	/* the target data (we actually use AEP_CMD_MSG_SIZE - overhead_size bytes,
									   so cmd is just slightly bigger than required which is OK) */
	AEP_BYTEBLOCK	 sCommandBlock;	/* the command message to send (header + target image) */
	AEP_MSG_HDR_PTR	 psMessageHdr;	/* Used for overlaying on top of buffered data */
	int bytes_written = 0;
	int bytes_read = 0;

	sCommandBlock.ptr = acCmd;

	GenerateMsgHeader( sCommandBlock.ptr,
					   AEP_C_SOFT_RESET,
					   0,
					   AEP_F_NOCTXCMD|AEP_F_SA,
					   0 );

	/* Set the total length of the message to just the header size as no data will be sent */
	sCommandBlock.Len = HDR_SIZE;

	/* Write to the destination device */
	bytes_written = WriteMessage( iDevFD, &sCommandBlock );

	if ( bytes_written < 0 ) {
		return RV_WRITE_ERROR;
	}

	/***************/
	/* Begin Read! */
	bytes_read = read(iDevFD, acReply, AEP_BLOCK_SIZE);

	if ( bytes_written != bytes_read ) {
		return RV_READ_ERROR;
	}

	/* Convert endiness of header */
	if( bytes_read >= HDR_SIZE )
		LEndToBEnd(acReply, HDR_SIZE/sizeof(AEP_U32));
	
	/* Point message header ptr to start of reply buffer */
	psMessageHdr=(AEP_MSG_HDR_PTR)acReply;

	/* Check that the message was received OK by the loader */
	if( psMessageHdr->status == AEP_S_COMPLETE ) {
		sleep(RESET_WAIT_TIME);
	}

	return RV_NO_ERROR;
}		 



/*******************************************************************************/
/*	BootTarget( ... )

	Downloads an image to the target.  The image is written sequentially to the 
	target starting at the load address.  A CRC is calculated which is then verified
	by the target and if orrect, the entry address given will be used to begin
	execution on the target.

	Input:	
	dev_fd		- The destination device file descriptor (target)
	img_fd		- Pointer to the image source file to be downloaded
	iTargetLoadAddress	- The first address on the target to download the image to
	iTargetEntryAddress - The address to begin execution on the target once d/l is complete

	Output:
	int			- Zero if Successful, negative if an error occurred

*/
int BootTarget( int dev_fd, FILE *img_fd, AEP_U32 iTargetLoadAddress, AEP_U32 iTargetEntryAddress )
{		
	int iBytesRead=0;				/* nb bytes read from a file */ 
	int iBytesWritten=0;			/* nb bytes written to a file */
	int keyb=0;						/* character read using getchar for pauses */
	int iCurrentLoadMsg=0;			/* message number when target image is split (when message size greater 
									   than AEP_CMD_MSG_SIZE) */ 
	int iOverheadLen=0;				/* size of message overhead = size of header + extra fields not including
									   target image data */
	long lImageLength=0;			/* real length of in bytes of the target to send */
	int iImageDataLengthPerMsg=0;	/* actual size of target image data that can be sent in a block of size
									   AEP_CMD_MSG_SIZE, value = AEP_CMD_MSG_SIZ - overhead_size */
	int iImageDataLength=0;			/* size of last target image data that was sent */
	AEP_U32 *piDataToSend;			/* Pointer to the next 32-bit value of non-image data to send */
	char *pcImageDataToSend;		/* Pointer to the start of the next image block to send */
	int iTargetMemOffset=0;			/* offset of data to write on target SDRAM, counting from
									   AEP_BOOT_IMG_ADDRESS */
	int iTargetCRC=startCRC16;		/* crc for the ENTIRE target image being loaded WITHOUT the message overhead
									   (see crc.h for explanations about this value) */

	/* Message buffers: */
	AEP_U8 acReply[AEP_BLOCK_SIZE]; /* the reply buffer (when target replies to a command) */
	AEP_U8 acCmd[AEP_BLOCK_SIZE];	/* the target data (we actually use AEP_CMD_MSG_SIZE - overhead_size bytes,
									   so cmd is just slightly bigger than required which is OK) */
	AEP_BYTEBLOCK	 sCommandBlock;	/* the command message to send (header + target image) */
	AEP_MSG_HDR_PTR	 psMessageHdr;	/* Used for overlaying on top of buffered data */
	AEP_BOOT_CMD_PTR psBootCmd;		/* Used for overlaying on top of buffered data */
	
	
	/**************** get image size *********************/
	/* Go to the end of the image file, then get the position of the stream to determine
	   the exact size of the file */
	fseek( img_fd, 0L, SEEK_END );
	lImageLength = ftell( img_fd );

	if ( lImageLength == 0 ) {
		/* The file is empty so exit with an error */
		fprintf( stderr, "ERROR: Image File %s is empty.\n", acAEPImageName );
		return RV_IMAGE_ERR;
	}
	
	/* now that we have the nb of AEP_U8s in the file, rewind back to the beginning */
	rewind(img_fd);	   
	
	/* Size of header, plus size of dest. address, the dest. address, and the length of the data */
	iOverheadLen = HDR_SIZE + 3 * sizeof(AEP_U32);
	iImageDataLengthPerMsg = AEP_BLOCK_SIZE - iOverheadLen;
	
	/* Init array/stucture pointers */
	sCommandBlock.ptr = acCmd;
	psMessageHdr = (AEP_MSG_HDR_PTR)acCmd;
	iCurrentLoadMsg = 0;
	
	while ( (signed int)( lImageLength - ( iCurrentLoadMsg * iImageDataLengthPerMsg )) > 0 ) {
		
		/**************** Send image message ***************/
		
		/* Set message length depending on bytes remaining */
		if ( lImageLength - ( iCurrentLoadMsg * iImageDataLengthPerMsg ) < iImageDataLengthPerMsg ) {
			/* Determine the final message length */
			iImageDataLength = lImageLength % iImageDataLengthPerMsg;
		}
		else {
			iImageDataLength = iImageDataLengthPerMsg;
		}
		
		if(bDoPause){	 
			fprintf(stderr, "\nHit ENTER to proceed with LOAD (write to device): ");
			keyb=getchar();
		}
		
		/* Initialize message header with load target command */
		GenerateMsgHeader( acCmd, 
						   AEP_C_LOAD_TARGET, 
						   0, 
						   AEP_F_NOCTXCMD|AEP_F_LOADER,
						   3*sizeof(AEP_U32)+iImageDataLength );
		
		/* Set int pointer to start of data array to insert dest. address parameter */
		piDataToSend = (AEP_U32 *)(acCmd + HDR_DATA_OFFSET);
		/* Set char pointer to address when the image data should start */
		pcImageDataToSend = acCmd + iOverheadLen;
		
		/* Destination address size = sizeof(AEP_U32) */
		WriteInt( piDataToSend++, sizeof(AEP_U32) );
		/* Destination address */
		WriteInt( piDataToSend++, iTargetLoadAddress + iTargetMemOffset );
		/* Size of image data to be transmitted */
		WriteInt( piDataToSend++, iImageDataLength );
		
		/* Read Image Data from file */
		iBytesRead = fread( pcImageDataToSend, sizeof(AEP_U8), iImageDataLength, img_fd);
		
		if( iBytesRead != iImageDataLength ){
			fprintf(stderr, "ERROR: Invalid Read from %s: %s\n",
					acAEPImageName, strerror(ferror(img_fd)));
			return RV_IMAGE_ERR;
		}
		
		/* compute crc of image data */
		iTargetCRC = crc16( sCommandBlock.ptr + iOverheadLen, iImageDataLength, iTargetCRC );
		
		/* Record the number of bytes sent in this message */
		sCommandBlock.Len = iOverheadLen + iImageDataLength;
		
		/* Write completed message to target, exit if there is a problem */
		if (( iBytesWritten = WriteMessage( dev_fd, &sCommandBlock )) < 0 ) {
			/* The number of bytes written was -1 so try to determine the cause
			of the problem */
			if ( errno == EAGAIN ) {
				/* The target has responded with a 'No Available resources',
				which, in response to a loader message, means that an application
				is currently loaded on the target. */
				return RV_APP_ALREADY_LOADED;
			}
			else {
				/* Report the unknown write error */
				return RV_WRITE_ERROR;
			}
		}
		
		/* Update offset on target memory to store image data consecutively on target */
		iTargetMemOffset += iImageDataLength;
		
		
		/**************** Read result ***************/
		
		if(bDoPause){
			fprintf(stderr, "\nHit ENTER to proceed with LOAD (read from device): ");
			keyb=getchar();
		}

		/* Read reply from target */
		iBytesRead = read(dev_fd, (AEP_U8*)acReply, AEP_BLOCK_SIZE);

		if(bVerbose) printf("\nGot 0x%08x (%d) AEP_U8\n", iBytesRead, iBytesRead);

		if ( iBytesWritten != iBytesRead ) {
			return RV_READ_ERROR;
		}
		
		if(iBytesRead>=HDR_SIZE) {
			/* Convert the reply header's endiness */
			LEndToBEnd((AEP_U8*)acReply, HDR_SIZE/sizeof(AEP_U32));
		}
		
		/* Map a message header over the reply buffer */
		psMessageHdr=(AEP_MSG_HDR_PTR)acReply;
		
		/* Check that the message was received OK by the loader */
		if((psMessageHdr->status != AEP_S_COMPLETE) || ( iBytesRead != iBytesWritten )){
			
			if(bVerbose) printf(" psMessageHdr->status =  0x%08x (%d)\n",  psMessageHdr->status, psMessageHdr->status);
			
			fprintf(stderr, "ERROR: Problem while sending image %s to accelerator device %s.\n",
					acAEPImageName, acAEPDeviceName );
			
			if(psMessageHdr->status == AEP_S_FXN_NOT_IMPLEMENTED) {
				fprintf(stderr, "Possible cause: The target has been previously loaded with an application.\n");
				return RV_APP_ALREADY_LOADED;
			}
			else {
				fprintf(stderr, "Please try to run current command again\n");
				return RV_TARGET_ERR;
			}
		}
		
		/* This message was completed successfully, move on to the next one */
		++iCurrentLoadMsg;
	}
	
	
	/******************** BOOT TARGET ***************************/
	
	if(bDoPause){	 
		fprintf(stderr, "hit ENTER to proceed with BOOT (write to device): ");
		keyb=getchar();
	}
	
	/************* generate header *************/
	
	if(bVerbose){	
		printf("\n************** BOOTING TARGET **************\n\n");	 
	}
	
	/* Initialize message header with boot target command */
	GenerateMsgHeader(acCmd, 
					  AEP_C_BOOT_TARGET,
					  0, 
					  AEP_F_NOCTXCMD|AEP_F_LOADER,
					  sizeof(AEP_U32)+AEP_BOOT_CMD_SIZE );
	
	/* Set int pointer to start of data array to insert dest. address parameter */
	piDataToSend = (AEP_U32 *)(acCmd + HDR_DATA_OFFSET);
	/* Set char pointer to address when the image data should start */
	pcImageDataToSend = acCmd + HDR_DATA_OFFSET + sizeof(AEP_U32);
	
	/************* Set data length *************/		 
	
	/* Boot command information size = AEP_BOOT_CMD_SIZE */
	WriteInt( piDataToSend++, AEP_BOOT_CMD_SIZE );
	
	/************* Set data *********************/
	
	psBootCmd = (AEP_BOOT_CMD_PTR)piDataToSend;
	
	psBootCmd->EntryAddress = iTargetEntryAddress;
	psBootCmd->StartAddress = iTargetLoadAddress;
	psBootCmd->TargetLength = (AEP_U32)lImageLength;
	psBootCmd->ExpectedCRC	= iTargetCRC;
	
	if(bVerbose) {
		printf(" psBootCmd->EntryAddress =	0x%08x (%d)\n",	psBootCmd->EntryAddress, psBootCmd->EntryAddress);
		printf(" psBootCmd->StartAddress =	0x%08x (%d)\n",	psBootCmd->StartAddress, psBootCmd->StartAddress);
		printf(" psBootCmd->TargetLength =	0x%08x (%d)\n",	psBootCmd->TargetLength, psBootCmd->TargetLength);
		printf(" psBootCmd->ExpectedCRC	 =	0x%08x (%d)\n",	psBootCmd->ExpectedCRC, psBootCmd->ExpectedCRC);
	}
	
	BEndToLEnd( (AEP_U8 *)psBootCmd, AEP_BOOT_CMD_SIZE/sizeof(AEP_U32) );
	
	if(bVerbose){
		printf("************************************\n");
		printf("Writing message to accelerator...");
		fflush(stdout);
	}
	
	/* Set block size: Header + length of boot info struct int + actual length of boot info struct */
	sCommandBlock.Len = HDR_SIZE + sizeof(AEP_U32) + AEP_BOOT_CMD_SIZE;

	iBytesWritten = write(dev_fd, sCommandBlock.ptr, sCommandBlock.Len);	
	
	if( iBytesWritten != sCommandBlock.Len ){
		return RV_WRITE_ERROR;
	}
	
	if(bDoPause){
		fprintf(stderr, "hit ENTER to proceed with BOOT (read from device):");
		keyb=getchar(); 
	}
	
	iBytesRead = read(dev_fd, (AEP_U8*)acReply, AEP_BLOCK_SIZE);  
	
	/* if booting is OK, we should get the same nb of bytes back: */
	if( iBytesRead != iBytesWritten ){	
		return RV_READ_ERROR;
	}
	
	if(bVerbose){
		printf("\nReceived 0x%08x (%d) AEP_U8\n", iBytesRead, iBytesRead);
		printf("converting endianness ..");
	}
	/* Convert endiness of boot message reply */
	LEndToBEnd((AEP_U8*)acReply, iBytesRead / sizeof(AEP_U32));
	if(bVerbose) printf("done\n");
	
	/* Init struct pointers to response */
	psMessageHdr = (AEP_MSG_HDR_PTR)acReply;	
	psBootCmd = (AEP_BOOT_CMD *)(acReply + HDR_SIZE + sizeof(AEP_U32));
	
	
	/* status set to AEP_S_COMPLETE if no crc error on target data */
	if((psMessageHdr->status != AEP_S_COMPLETE) || (psBootCmd->ExpectedCRC != iTargetCRC )) {
		if(bVerbose) {
			printf(" psBootCmd->EntryAddress =	0x%08x (%d)\n",	psBootCmd->EntryAddress, psBootCmd->EntryAddress);
			printf(" psBootCmd->StartAddress =	0x%08x (%d)\n",	psBootCmd->StartAddress, psBootCmd->StartAddress);
			printf(" psBootCmd->TargetLength =	0x%08x (%d)\n",	psBootCmd->TargetLength, psBootCmd->TargetLength);
			printf(" psBootCmd->ExpectedCRC	 =	0x%08x (%d)\n",	psBootCmd->ExpectedCRC, psBootCmd->ExpectedCRC);
		}
		fprintf(stderr, "ERROR: Incorrect CRC while sending image %s to accelerator device %s.\n",
				acAEPImageName, acAEPDeviceName );
		fprintf(stderr, "Possible cause: Faulty accelerator card or bus error.\n");
		fprintf(stderr, "Please try to run current command again\n");
		return RV_CRC_INCORRECT; 
	}
	
	return 0;
}


/*******************************************************************************/
/*	usage()

	Prints the parameter usage info for this application to the user.

	Input:	
	None
	Output:	
	Text to stdout
	int		RV_MISSING_ARGS, signalling that the download should not continue because
			the input parameters are incorrect.

*/
int usage(void)
{
	printf("\nusage: aepload -device <device name> -image <target binary file name>\n\n");
	printf("(REQUIRED)	-device <Accelerator device name> = (eg: /dev/paep0, /dev/paep1, etc)\n");
	printf("(REQUIRED)	-image	<Target binary file name> = (eg: aeptarg.bin )\n");
	printf("(OPTIONAL)	-load	<Destination address for application [in hex]> (DEFAULT = 40000)\n");
	printf("(OPTIONAL)	-entry	<Entry point in application [in hex]>          (DEFAULT = 40000)\n");
	printf("(OPTIONAL)	-reset	 : Send reset command before executing download\n");
	printf("(OPTIONAL)	-version : Display version number and exit\n");
#ifdef THIS_IS_NOT_FOR_RELEASE /* obviously, the customer should not have access to this: */
	printf("(OPTIONAL)	-verbose : enable verbose output	  (DEFAULT = disabled)\n");
	printf("(OPTIONAL)	-pause	 : pause between reads and writes (DEFAULT = disabled)\n\n");
#endif	  
	return RV_MISSING_ARGS;
}
